# -*- coding: UTF-8 -*-
#

# Learn more:
#
# https://support.testrail.com/hc/en-us/categories/7076541806228-API-Manual
#
# Copyright Gurock Software GmbH. See license.md for details.
#


import requests
import time
from urllib.parse import urljoin


class APIClient:
    def __init__(self, base_url, user, password, **kwargs):  # type: ignore
        """
        Instantiate the APIClient class.

        :param base_url: The same TestRail address for the API client you also use to access TestRail with your web
            browser (e.g., https://<your-name>.testrail.com/ or http://<server>/testrail/).
        :type base_url: Str
        :param user: Username for the account on the TestRail server.
        :type user: Str
        :param password: Password for the account on the TestRail server.
        :type password: Str
        :param headers: (optional) Dictionary of HTTP Headers to send with each request.
        :type headers: Dict
        :param cert_check: (optional) Either a boolean, in which case it controls whether we verify the server's TLS
            certificate, or a string, in which case it must be a path to a CA bundle to use. Defaults to ``True``.
        :type cert_check: Bool or str
        :param timeout: (optional) How many seconds to wait for the server to send data before giving up, as a float,
            or a: ref: `(connect timeout, read timeout) <timeouts>` tuple.
        :type timeout: Float or tuple
        """
        self.user = user
        self.password = password
        self._url = urljoin(base_url, "index.php?/api/v2/")
        self.headers = kwargs.get("headers", {"Content-Type": "application/json"})
        self.cert_check = kwargs.get("cert_check", True)
        self.timeout = kwargs.get("timeout", 10.0)
        if self.timeout is not None:
            self.timeout = (
                isinstance(self.timeout, float) if False else float(self.timeout)
            )

    def send_get(self, uri, **kwargs):  # type: ignore
        """
        Send GET

        Issues a GET request (read) against the API and returns the result (as Python dict).

        :param uri: The API method to call including parameters (e.g., get_case/1)
        :type uri: str
        :param headers: (optional) Dictionary of HTTP Headers to send with the request.
        :type headers: dict
        :param cert_check: (optional) Either a boolean, in which case it controls whether we verify the server's TLS
            certificate, or a string, in which case it must be a path to a CA bundle to use. Defaults to ``True``.
        :type cert_check: bool or str
        :param timeout: (optional) How many seconds to wait for the server to send data before giving up, as a float,
            or a: ref: `(connect timeout, read timeout) <timeouts>` tuple.
        :type timeout: float or tuple
        """
        cert_check = kwargs.get("cert_check", self.cert_check)
        headers = kwargs.get("headers", self.headers)
        url = self._url + uri
        r = requests.get(
            url,
            auth=(self.user, self.password),
            headers=headers,
            verify=cert_check,
            timeout=self.timeout,
        )

        if r.status_code == 429:  # Too many requests
            pause = int(r.headers.get("Retry-After", 60))
            print("Too many requests: pause for {}s".format(pause))
            time.sleep(pause)
            return self.send_get(uri, **kwargs)
        else:
            return r.json()

    def send_post(self, uri, data, **kwargs):  # type: ignore
        """
        Send POST

        Issues a POST request (write) against the API and returns the result (as Python dict).

        :param uri: The API method to call including parameters (e.g., get_case/1) is
        :type uri: str
        :param data: The data to submit as part of the request (strings must be UTF-8 encoded).
        :type data: Dict
        :param headers: (optional) Dictionary of HTTP Headers to send with the request.
        :type headers: Dict
        :param cert_check: (optional) Either a boolean, in which case it controls whether we verify the server's TLS
            certificate, or a string, in which case it must be a path to a CA bundle to use. Defaults to ``True``.
        :type cert_check: Bool or str
        :param timeout: (optional) How many seconds to wait for the server to send data before giving up, as a float,
            or a: ref: `(connect timeout, read timeout) <timeouts>` tuple.
        :type timeout: Float or tuple
        """
        cert_check = kwargs.get("cert_check", self.cert_check)
        headers = kwargs.get("headers", self.headers)
        url = self._url + uri
        r = requests.post(
            url,
            auth=(self.user, self.password),
            headers=headers,
            json=data,
            verify=cert_check,
            timeout=self.timeout,
        )

        if r.status_code == 429:  # Too many requests
            pause = int(r.headers.get("Retry-After", 60))
            print("Too many requests: pause for {}s".format(pause))
            time.sleep(pause)
            return self.send_post(uri, data, **kwargs)
        else:
            return r.json()

    @staticmethod
    def get_error(json_response):  # type: ignore
        """Extract error contained in an API response.
        If no error occurred, return None

        :param json_response: json response of request
        :return: String of the error
        """
        if "error" in json_response and json_response["error"]:
            return json_response["error"]



# -*- coding: UTF-8 -*-
import os
from typing import Union, Optional
from .plugin import PyTestRailPlugin
from .testrail_api import APIClient
from _pytest.config import Config
from _pytest.config.argparsing import Parser
import configparser


def pytest_addoption(parser: Parser) -> None:
    group = parser.getgroup("testrail")
    group.addoption(
        "--testrail",
        action="store_true",
        help="Create and update testruns with TestRail",
    )
    group.addoption(
        "--tr-config",
        action="store",
        default="testrail.cfg",
        help="Path to the config file containing information about the TestRail server (defaults to testrail.cfg)",
    )
    group.addoption(
        "--tr-url",
        action="store",
        help="TestRail address you use to access TestRail with your web browser (config file: url in API section)",
    )
    group.addoption(
        "--tr-email",
        action="store",
        help="Email for the account on the TestRail server (config file: email in API section)",
    )
    group.addoption(
        "--tr-password",
        action="store",
        help="Password for the account on the TestRail server (config file: password in API section)",
    )
    group.addoption(
        "--tr-timeout",
        action="store",
        help="Set timeout for connecting to TestRail server",
    )
    group.addoption(
        "--tr-testrun-assignedto-id",
        action="store",
        help="ID of the user assigned to the test run (config file: assignedto_id in TESTRUN section)",
    )
    group.addoption(
        "--tr-testrun-project-id",
        action="store",
        help="ID of the project the test run is in (config file: project_id in TESTRUN section)",
    )
    group.addoption(
        "--tr-testrun-suite-id",
        action="store",
        help="ID of the test suite containing the test cases (config file: suite_id in TESTRUN section)",
    )
    group.addoption(
        "--tr-testrun-suite-include-all",
        action="store_true",
        default=None,
        help="Include all test cases in specified test suite when creating test run (config file: include_all in TESTRUN section)",
    )
    group.addoption(
        "--tr-testrun-name",
        action="store",
        default=None,
        help="Name given to testrun, that appears in TestRail (config file: name in TESTRUN section)",
    )
    group.addoption(
        "--tr-testrun-description",
        action="store",
        default=None,
        help="Description given to testrun, that appears in TestRail (config file: description in TESTRUN section)",
    )
    group.addoption(
        "--tr-run-id",
        action="store",
        default=0,
        required=False,
        help='Identifier of testrun, that appears in TestRail. If provided, option "--tr-testrun-name" will be ignored',
    )
    group.addoption(
        "--tr-plan-id",
        action="store",
        required=False,
        help='Identifier of testplan, that appears in TestRail (config file: plan_id in TESTRUN section). If provided, option "--tr-testrun-name" will be ignored',
    )
    group.addoption(
        "--tr-version",
        action="store",
        default="",
        required=False,
        help="Indicate a version in Test Case result",
    )
    group.addoption(
        "--tr-no-ssl-cert-check",
        action="store_false",
        default=None,
        help="Do not check for valid SSL certificate on TestRail host",
    )
    group.addoption(
        "--tr-close-on-complete",
        action="store_true",
        default=False,
        required=False,
        help="Close a test run on completion",
    )
    group.addoption(
        "--tr-dont-publish-blocked",
        action="store_false",
        required=False,
        help='Determine if results of "blocked" testcases (in TestRail) are published or not',
    )
    group.addoption(
        "--tr-skip-missing",
        action="store_true",
        required=False,
        help="Skip test cases that are not present in testrun",
    )
    group.addoption(
        "--tr-milestone-id",
        action="store",
        default=None,
        required=False,
        help="Identifier of milestone, to be used in run creation (config file: milestone_id in TESTRUN section)",
    )
    group.addoption(
        "--tc-custom-comment",
        action="store",
        default=None,
        required=False,
        help="Custom comment, to be appended to default comment for test case (config file: custom_comment in TESTCASE section)",
    )


def pytest_configure(config: Config) -> None:
    if config.getoption("--testrail"):
        cfg_file_path = config.getoption("--tr-config")
        config_manager = ConfigManager(cfg_file_path, config)
        client = APIClient(
            config_manager.getoption("tr-url", "url", "API"),
            config_manager.getoption("tr-email", "email", "API"),
            config_manager.getoption("tr-password", "password", "API"),
            timeout=config_manager.getoption("tr-timeout", "timeout", "API"),
        )

        config.pluginmanager.register(
            PyTestRailPlugin(
                client=client,
                assign_user_id=int(
                    config_manager.getoption(
                        "tr-testrun-assignedto-id", "assignedto_id", "TESTRUN"
                    )
                    or 0
                ),
                project_id=int(
                    config_manager.getoption(
                        "tr-testrun-project-id", "project_id", "TESTRUN"
                    )
                    or 0
                ),
                suite_id=int(
                    config_manager.getoption(
                        "tr-testrun-suite-id", "suite_id", "TESTRUN"
                    )
                    or 0
                ),
                include_all=bool(
                    config_manager.getoption(
                        "tr-testrun-suite-include-all",
                        "include_all",
                        "TESTRUN",
                        is_bool=True,
                        default=False,
                    )
                ),
                cert_check=bool(
                    config_manager.getoption(
                        "tr-no-ssl-cert-check",
                        "no_ssl_cert_check",
                        "API",
                        is_bool=True,
                        default=True,
                    )
                ),
                tr_name=str(
                    config_manager.getoption("tr-testrun-name", "name", "TESTRUN") or ""
                ),
                tr_description=str(
                    config_manager.getoption(
                        "tr-testrun-description", "description", "TESTRUN"
                    )
                    or ""
                ),
                run_id=config.getoption("--tr-run-id"),
                plan_id=int(
                    config_manager.getoption("tr-plan-id", "plan_id", "TESTRUN") or 0
                ),
                version=config.getoption("--tr-version"),
                close_on_complete=config.getoption("--tr-close-on-complete"),
                publish_blocked=config.getoption("--tr-dont-publish-blocked"),
                skip_missing=config.getoption("--tr-skip-missing"),
                milestone_id=int(
                    config_manager.getoption(
                        "tr-milestone-id", "milestone_id", "TESTRUN"
                    )
                    or 0
                ),
                custom_comment=str(
                    config_manager.getoption(
                        "tc-custom-comment", "custom_comment", "TESTCASE"
                    )
                    or ""
                ),
            ),
            # Name of plugin instance (allow to be used by other plugins)
            name="pytest-testrail-instance",
        )


class ConfigManager(object):
    def __init__(self, cfg_file_path: Optional[str], config: Config) -> None:
        """
        Handles retrieving configuration values. Config options set in flags are given preferance over options set in the
        config file.

        :param cfg_file_path: Path to the config file containing information about the TestRail server.
        :type cfg_file_path: str or None
        :param config: Config object containing commandline flag options.
        :type config: _pytest.config.Config
        """
        self.cfg_file = None
        if cfg_file_path is not None and (
            os.path.isfile(cfg_file_path) or os.path.islink(cfg_file_path)
        ):
            self.cfg_file = configparser.ConfigParser()
            self.cfg_file.read(cfg_file_path)

        self.config = config

    def getoption(
        self,
        flag: str,
        cfg_name: str,
        section: Optional[str] = None,
        is_bool: bool = False,
        default: Optional[Union[str, bool]] = None,
    ) -> Optional[Union[str, bool]]:
        # priority: cli > config file > default

        # 1. return cli option (if set)
        value = self.config.getoption("--{}".format(flag))
        if value is not None:
            return value

        # 2. return default if not config file path is specified
        if section is None or self.cfg_file is None:
            return default

        if self.cfg_file.has_option(section, cfg_name):
            # 3. return config file value
            return (
                self.cfg_file.getboolean(section, cfg_name)
                if is_bool
                else self.cfg_file.get(section, cfg_name)
            )
        else:
            # 4. if entry not found in config file
            return default



# -*- coding: UTF-8 -*-
from datetime import datetime
from operator import itemgetter
from typing import Tuple, Optional, Any

import _pytest
from _pytest.nodes import Node
from _pytest.mark.structures import MarkDecorator
from _pytest.python import Function
from _pytest.config import Config
from _pytest.main import Session
from _pytest.runner import CallInfo
import pytest
import re
import warnings

from .testrail_api import APIClient

import logging

# Turn on if you want to see the logs

# Create a folder for logs if it does not already exist
# if "BITBUCKET_CLONE_DIR" in os.environ:
#     log_directory = os.path.join(os.environ["BITBUCKET_CLONE_DIR"], "logs")
# else:
#     log_directory = os.path.abspath("test_reports/logs/")
#
# if not os.path.exists(log_directory):
#     os.makedirs(log_directory)
#
# # Log file name
# log_file_name = os.path.join(log_directory, "testrail_plugin.log")
#
# # Logger setup
# logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)
#
# # Formatter for logs
# formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
#
# # Handler for writing logs to a file
# file_handler = logging.FileHandler(log_file_name)
# file_handler.setFormatter(formatter)
# logger.addHandler(file_handler)


# Logger setup
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

# Formatter for logs
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")

# # Handler for writing logs to a file
# file_handler = logging.FileHandler(log_file_name)
# file_handler.setFormatter(formatter)
# logger.addHandler(file_handler)


CUSTOM_TESTRAIL_STATUS = {
    "passed": 1,
    "blocked": 2,
    "untested": 3,
    "retest": 4,
    "failed": 5,
    "xfailed": 5,  # Mark expected failures as failed
    "xpassed": 1,  # Mark unexpected passes as passed
}

PYTEST_TO_TESTRAIL_STATUS = {
    "passed": CUSTOM_TESTRAIL_STATUS["passed"],
    "failed": CUSTOM_TESTRAIL_STATUS["failed"],
    "skipped": CUSTOM_TESTRAIL_STATUS["blocked"],
    "xpassed": CUSTOM_TESTRAIL_STATUS["passed"],
    "xfailed": CUSTOM_TESTRAIL_STATUS["failed"],
}

DT_FORMAT = "%d-%m-%Y %H:%M:%S"

TESTRAIL_PREFIX = "testrail"
TESTRAIL_DEFECTS_PREFIX = "testrail_defects"
ADD_RESULTS_URL = "add_results_for_cases/{}"
ADD_TESTRUN_URL = "add_run/{}"
CLOSE_TESTRUN_URL = "close_run/{}"
CLOSE_TESTPLAN_URL = "close_plan/{}"
GET_TESTRUN_URL = "get_run/{}"
GET_TESTPLAN_URL = "get_plan/{}"
GET_TESTS_URL = "get_tests/{}"

COMMENT_SIZE_LIMIT = 4000


class DeprecatedTestDecorator(DeprecationWarning):
    pass


warnings.simplefilter(action="once", category=DeprecatedTestDecorator, lineno=0)


class pytestrail(object):
    """
    An alternative to using the testrail function as a decorator for test cases, since pytest may confuse it as a test
    function since it has the 'test' prefix
    """

    @staticmethod
    def case(*ids: str) -> MarkDecorator:
        """
        Decorator to mark tests with testcase ids.

        i.e., @pytestrail.case('C123', 'C12345')

        :return pytest.mark:
        """
        return pytest.mark.testrail(ids=ids)

    @staticmethod
    def defect(*defect_ids: str) -> MarkDecorator:
        """
        Decorator to mark defects with defect ids.

        i.e. @pytestrail.defect('PF-513', 'BR-3255')

        :return pytest.mark:
        """
        return pytest.mark.testrail_defects(defect_ids=defect_ids)


def testrail(*ids: str) -> MarkDecorator:
    """
    Decorator to mark tests with testcase ids.

    ie. @testrail('C123', 'C12345')

    :return pytest.mark:
    """
    deprecation_msg = (
        "pytest_testrail: the @testrail decorator is deprecated and will be removed. Please use the "
        "@pytestrail.case decorator instead."
    )
    warnings.warn(deprecation_msg, DeprecatedTestDecorator)
    return pytestrail.case(*ids)


def get_test_outcome(outcome: str) -> int:
    """
    Return numerical value of a test outcome.

    :param str outcome: pytest reported test outcome value.
    :returns: int relating to a test outcome.
    """
    if outcome == "xfailed":
        logger.debug(
            f"Test outcome 'xfailed' is mapped to TestRail status {CUSTOM_TESTRAIL_STATUS['failed']}"
        )
        return CUSTOM_TESTRAIL_STATUS["failed"]
    elif outcome == "xpassed":
        logger.debug(
            f"Test outcome 'xpassed' is mapped to TestRail status {CUSTOM_TESTRAIL_STATUS['passed']}"
        )
        return CUSTOM_TESTRAIL_STATUS["passed"]
    logger.debug(
        f"Test outcome '{outcome}' is mapped to TestRail status {PYTEST_TO_TESTRAIL_STATUS[outcome]}"
    )
    return PYTEST_TO_TESTRAIL_STATUS[outcome]


def testrun_name() -> str:
    """Returns testrun name with timestamp"""
    now = datetime.utcnow()
    return "Automated Run {}".format(now.strftime(DT_FORMAT))


def clean_test_ids(test_ids: list[str]) -> list[int]:
    """
    Clean pytest marker containing testrail testcase ids.

    :param list test_ids: list of test_ids.
    :return list ints: contains a list of test_ids as ints.
    """
    return [
        int(match.groupdict()["test_id"])
        for test_id in test_ids
        if (match := re.search("(?P<test_id>[0-9]+$)", test_id)) is not None
        and "test_id" in match.groupdict()
    ]


def clean_test_defects(defect_ids: list[str]) -> list[str]:
    """
    Clean pytest marker containing testrail defects ids.

    :param list defect_ids: list of defect_ids.
    :return list str: contains a list of defect_ids as str.
    """
    return [
        match.groupdict()["defect_id"]
        for defect_id in defect_ids
        if (match := re.search("(?P<defect_id>.*)", defect_id)) is not None
        and "defect_id" in match.groupdict()
    ]


def get_testrail_keys(items: list[Node]) -> list[Tuple[Node, list[int]]]:
    """Return Tuple of Pytest nodes and TestRail ids from pytests markers"""
    testcaseids = []
    for item in items:
        marker = item.get_closest_marker(TESTRAIL_PREFIX)
        if marker is not None and hasattr(marker, "kwargs"):
            ids = marker.kwargs.get("ids")
            if ids is not None and isinstance(ids, list):
                testcaseids.append(
                    (
                        item,
                        clean_test_ids(ids),
                    )
                )
    return testcaseids


class PyTestRailPlugin(object):
    def __init__(
        self,
        client: APIClient,
        assign_user_id: int,
        project_id: int,
        suite_id: int,
        include_all: bool,
        cert_check: bool,
        tr_name: str,
        tr_description: str = "",
        run_id: int = 0,
        plan_id: int = 0,
        version: str = "",
        close_on_complete: bool = False,
        publish_blocked: bool = True,
        skip_missing: bool = False,
        milestone_id: Optional[int] = None,
        custom_comment: Optional[str] = None,
    ) -> None:
        self.assign_user_id = assign_user_id
        self.cert_check = cert_check
        self.client = client
        self.project_id = project_id
        self.results: list[dict[str, Any]] = []
        self.suite_id = suite_id
        self.include_all = include_all
        self.testrun_name = tr_name
        self.testrun_description = tr_description
        self.testrun_id = run_id
        self.testplan_id = plan_id
        self.version = version
        self.close_on_complete = close_on_complete
        self.publish_blocked = publish_blocked
        self.skip_missing = skip_missing
        self.milestone_id = milestone_id
        self.custom_comment = custom_comment
        self.need_new_testrun = False
        logger.debug("PyTestRailPlugin initialized")
        # pytest hooks

    def pytest_report_header(self, config: _pytest.config.Config, startdir: str) -> str:
        """Add extra-info in header"""
        logger.debug("Creating report header")
        message = "pytest-testrail: "
        if self.testplan_id:
            message += "existing testplan #{} selected".format(self.testplan_id)
        elif self.testrun_id:
            message += "existing testrun #{} selected".format(self.testrun_id)
        else:
            message += "a new testrun will be created"
        logger.debug("Report header created")
        return message

    @pytest.hookimpl(trylast=True)
    def pytest_collection_modifyitems(
        self, session: Session, config: Config, items: list[Node]
    ) -> None:
        logger.debug("Modifying collected items")
        items_with_tr_keys = get_testrail_keys(items)
        tr_keys = [case_id for item in items_with_tr_keys for case_id in item[1]]

        if self.testplan_id:
            if not self.is_testplan_available():
                self.create_test_run(
                    self.assign_user_id,
                    self.project_id,
                    self.suite_id,
                    self.include_all,
                    self.testrun_name,
                    tr_keys,
                    self.milestone_id,
                    self.testrun_description,
                )
        elif self.testrun_id:
            if not self.is_testrun_available():
                self.create_test_run(
                    self.assign_user_id,
                    self.project_id,
                    self.suite_id,
                    self.include_all,
                    self.testrun_name,
                    tr_keys,
                    self.milestone_id,
                    self.testrun_description,
                )
            elif self.skip_missing:
                tests = self.get_tests(self.testrun_id)
                if tests is not None:
                    tests_list = [test.get("case_id") for test in tests]
                else:
                    tests_list = []
                for item, case_id in items_with_tr_keys:
                    if not set(case_id).intersection(set(tests_list)):
                        mark = pytest.mark.skip("Test is not present in testrun.")
                        item.add_marker(mark)
        else:
            self.create_test_run(
                self.assign_user_id,
                self.project_id,
                self.suite_id,
                self.include_all,
                self.testrun_name,
                tr_keys,
                self.milestone_id,
                self.testrun_description,
            )

    @pytest.hookimpl(tryfirst=True, hookwrapper=True)
    def pytest_runtest_makereport(self, item: Function, call: CallInfo) -> Any:
        """Collect result and associated testcases (TestRail) of an execution"""
        logger.info(f"Creating a test report for {item.name}")
        logger.debug(f"Creating report for test {item.name}")
        outcome = yield
        rep = outcome.get_result()

        # Logging the state of the rep object
        logger.debug(
            f"Test {item.name}: outcome={rep.outcome}, when={rep.when}, failed={rep.failed}, skipped={rep.skipped}, passed={rep.passed}"
        )

        # Check for xfail and change rep.outcome accordingly
        if "xfail" in item.keywords:
            logger.debug(f"Test {item.name} is marked as xfail")
            if rep.when == "call" and (rep.failed or rep.skipped):
                rep.outcome = "failed"
                logger.debug(f"Test {item.name} xfailed and marked as failed")
            elif rep.passed:
                rep.outcome = "passed"
                logger.debug(f"Test {item.name} xpassed")
            else:
                xfail_marker = item.get_closest_marker("xfail")
                if (
                    xfail_marker is not None
                    and xfail_marker.kwargs.get("run", True) == False
                ):
                    rep.outcome = "xfailed"  # Map 'xfail_run_false' to 'xfailed'
                    logger.debug(f"Test {item.name} is marked as xfail with run=False")

        logger.info(f"Completed processing test report for {item.name}")
        logger.debug(f"After xfail processing: {item.name} - Outcome: {rep.outcome}")

        # Add results to TestRail
        defectids = None
        test_parametrize = None
        if "callspec" in dir(item):
            test_parametrize = (
                list(item.callspec.params.values()) if item.callspec.params else None
            )
        comment = rep.longrepr
        defect_marker = item.get_closest_marker(TESTRAIL_DEFECTS_PREFIX)
        if defect_marker is not None:
            defectids = defect_marker.kwargs.get("defect_ids")
        test_marker = item.get_closest_marker(TESTRAIL_PREFIX)
        if test_marker is not None:
            testcaseids = test_marker.kwargs.get("ids")
            if testcaseids is not None:
                # Add results for both 'call' and 'setup' stages
                if rep.when == "call" or (
                    rep.when == "setup"
                    and (rep.outcome == "skipped" or rep.outcome == "xfailed")
                ):
                    if defectids is not None:
                        self.add_result(
                            item,
                            clean_test_ids(testcaseids),
                            rep,
                            comment=comment,
                            duration=rep.duration,
                            defects=str(clean_test_defects(defectids))
                            .replace("[", "")
                            .replace("]", "")
                            .replace("'", ""),
                            test_parametrize=test_parametrize,
                        )
                    else:
                        self.add_result(
                            item,
                            clean_test_ids(testcaseids),
                            rep,
                            comment=comment,
                            duration=rep.duration,
                            test_parametrize=test_parametrize,
                        )

    def pytest_sessionfinish(self, session: Session, exitstatus: int) -> None:
        """Publish results in TestRail"""
        logger.debug("Session finished, publishing results")
        print("[{}] Start publishing".format(TESTRAIL_PREFIX))
        if self.results:
            tests_list = [str(result["case_id"]) for result in self.results]
            print(
                "[{}] Testcases to publish: {}".format(
                    TESTRAIL_PREFIX, ", ".join(tests_list)
                )
            )

            if self.testrun_id:
                self.add_results(self.testrun_id)
            elif self.testplan_id:
                testruns = self.get_available_testruns(self.testplan_id)
                print(
                    "[{}] Testruns to update: {}".format(
                        TESTRAIL_PREFIX, ", ".join([str(elt) for elt in testruns])
                    )
                )
                for testrun_id in testruns:
                    self.add_results(testrun_id)
            else:
                print("[{}] No data published".format(TESTRAIL_PREFIX))

            if self.close_on_complete and self.testrun_id:
                self.close_test_run(self.testrun_id)
            elif self.close_on_complete and self.testplan_id:
                self.close_test_plan(self.testplan_id)
        print("[{}] End publishing".format(TESTRAIL_PREFIX))

    def add_result(
        self,
        item: _pytest.python.Function,
        test_ids: list[int],
        rep: _pytest.runner.TestReport,
        comment: str = "",
        defects: Optional[str] = None,
        duration: float = 0,
        test_parametrize: Optional[list[Any]] = None,
    ) -> None:
        """
        Add a new result to result dict to be submitted at the end.

        :param item: The pytest Function object representing the test item.
        :param list test_parametrize: Add test parametrize to test result
        :param defects: Add defects to a test result
        :param list test_ids: list of test_ids.
        :param _pytest.runner.TestReport rep: Pytest report object with test result details.
        :param comment: None or a failure representation.
        :param duration: Time it took to run just the test.
        """
        logger.debug(f"Processing test result, initial outcome: {rep.outcome}")
        for test_id in test_ids:
            xfail_marker = item.get_closest_marker("xfail")
            if xfail_marker is not None and xfail_marker.kwargs.get("run", True):
                # If the test is marked as xfail and run=True
                status = (
                    CUSTOM_TESTRAIL_STATUS["failed"]
                    if rep.wasxfail
                    else CUSTOM_TESTRAIL_STATUS["blocked"]
                )
            else:
                # If the test is marked as xfail and run=False
                status = CUSTOM_TESTRAIL_STATUS["failed"]
            # For all other cases, use the general mapping
            status = get_test_outcome(rep.outcome)

            # Logging before adding result
            logger.debug(
                f"Adding result for test {test_id}: status={status}, outcome={rep.outcome}"
            )

            # Create a result record
            data = {
                "case_id": test_id,
                "status_id": status,
                "comment": comment,
                "duration": duration,
                "defects": defects,
                "test_parametrize": test_parametrize,
            }
            self.results.append(data)

        logger.debug("Test result added")

    def add_results(self, testrun_id: int) -> None:
        """
        Add results one by one to improve error handling.

        :param testrun_id: ID of the testrun to feed
        """
        logger.info(f"Adding results for testrun_id: {testrun_id}")
        logger.debug(f"Adding results for testrun_id: {testrun_id}")
        # Results are sorted by 'case_id' and by 'status_id' (the worst result at the end)
        self.results.sort(key=itemgetter("case_id"))

        # Manage a case of "blocked" testcases
        if self.publish_blocked is False:
            print(
                '[{}] Option "Don\'t publish blocked testcases" activated'.format(
                    TESTRAIL_PREFIX
                )
            )
            tests = self.get_tests(testrun_id)
            if tests is not None:
                blocked_tests_list = [
                    test.get("case_id")
                    for test in tests
                    if test.get("status_id") == CUSTOM_TESTRAIL_STATUS["blocked"]
                ]
            else:
                blocked_tests_list = []
            print(
                "[{}] Blocked testcases excluded: {}".format(
                    TESTRAIL_PREFIX, ", ".join(str(elt) for elt in blocked_tests_list)
                )
            )
            self.results = [
                result
                for result in self.results
                if result.get("case_id") not in blocked_tests_list
            ]

        # prompt enabling include all test cases from the test suite when creating test run
        if self.include_all:
            print(
                '[{}] Option "Include all testcases from test suite for test run" activated'.format(
                    TESTRAIL_PREFIX
                )
            )

        # Publish results
        data: dict[str, Any] = {"results": []}
        for result in self.results:
            logger.debug(
                f"Preparing to send result for case {result['case_id']} with status {result['status_id']}"
            )
            entry = {
                "status_id": result["status_id"],
                "case_id": result["case_id"],
                "defects": result["defects"],
            }
            if self.version:
                entry["version"] = self.version
            comment = result.get("comment", "")
            test_parametrize = result.get("test_parametrize", "")
            entry["comment"] = ""
            if test_parametrize:
                entry["comment"] += "# Test parametrize: #\n"
                entry["comment"] += str(test_parametrize) + "\n\n"
            if comment:
                if self.custom_comment:
                    entry["comment"] += self.custom_comment + "\n"
                    # Indent text to avoid string formatting by TestRail. Limit the size of comment.
                    entry["comment"] += "# Pytest result: #\n"
                    entry["comment"] += (
                        "Log truncated\n...\n"
                        if len(str(comment)) > COMMENT_SIZE_LIMIT
                        else ""
                    )
                    entry["comment"] += "    " + str(comment)[
                        -COMMENT_SIZE_LIMIT:
                    ].replace("\n", "\n    ")
                else:
                    # Indent text to avoid string formatting by TestRail. Limit the size of comment.
                    entry["comment"] += "# Pytest result: #\n"
                    entry["comment"] += (
                        "Log truncated\n...\n"
                        if len(str(comment)) > COMMENT_SIZE_LIMIT
                        else ""
                    )
                    entry["comment"] += "    " + str(comment)[
                        -COMMENT_SIZE_LIMIT:
                    ].replace("\n", "\n    ")
            elif comment == "":
                entry["comment"] = self.custom_comment
            duration = result.get("duration")
            if duration:
                duration = (
                    1 if (duration < 1) else int(round(duration))
                )  # TestRail API doesn't manage milliseconds
                entry["elapsed"] = str(duration) + "s"
            data["results"].append(entry)

        response = self.client.send_post(
            ADD_RESULTS_URL.format(testrun_id), data, cert_check=self.cert_check
        )
        error = self.client.get_error(response)
        if error:
            logger.error(f"Failed to add results: {error}")
            print(
                '[{}] Info: Testcases not published for following reason: "{}"'.format(
                    TESTRAIL_PREFIX, error
                )
            )
        else:
            logger.info("Results added successfully")

    def create_test_run(
        self,
        assign_user_id: int,
        project_id: int,
        suite_id: int,
        include_all: bool,
        testrun_name: str,
        tr_keys: list[int],
        milestone_id: Optional[int],
        description: str = "",
    ) -> None:
        """
        Create a test run with IDs collected from markers.

        :param assign_user_id: The ID of the user to whom the test run is assigned.
        :param project_id: The ID of the project in which the test run is created.
        :param suite_id: The ID of the test suite for the test run.
        :param include_all: Boolean indicating whether to include all test cases in the test run.
        :param testrun_name: The name of the test run.
        :param tr_keys: Collected TestRail IDs.
        :param milestone_id: The ID of the milestone associated with the test run. Optional.
        :param description: The description of the test run. Optional.
        """
        logger.info(f"Creating a test run named {testrun_name}")
        data = {
            "suite_id": suite_id,
            "name": testrun_name,
            "description": description,
            "assignedto_id": assign_user_id,
            "include_all": include_all,
            "case_ids": tr_keys,
            "milestone_id": milestone_id,
        }

        try:
            response = self.client.send_post(
                ADD_TESTRUN_URL.format(project_id), data, cert_check=self.cert_check
            )
            error = self.client.get_error(response)
            if error:
                # Log an error message if there is a problem with creating the test run
                logger.error(f'[TestRail Plugin] Error creating test run: "{error}"')
            else:
                # Log a success message with the new test run ID
                self.testrun_id = response["id"]
                logger.info(
                    f"[TestRail Plugin] Test run created successfully with ID={self.testrun_id}"
                )
        except Exception as e:
            # Log any unexpected exceptions during the test run creation
            logger.error(
                f"[TestRail Plugin] Unexpected error during test run creation: {e}"
            )

    def close_test_run(self, testrun_id: int) -> None:
        """
        Closes testrun.

        """
        response = self.client.send_post(
            CLOSE_TESTRUN_URL.format(testrun_id), data={}, cert_check=self.cert_check
        )
        error = self.client.get_error(response)
        if error:
            print('[{}] Failed to close test run: "{}"'.format(TESTRAIL_PREFIX, error))
        else:
            print(
                "[{}] Test run with ID={} was closed".format(
                    TESTRAIL_PREFIX, self.testrun_id
                )
            )

    def close_test_plan(self, testplan_id: int) -> None:
        """
        Closes testrun.

        """
        response = self.client.send_post(
            CLOSE_TESTPLAN_URL.format(testplan_id), data={}, cert_check=self.cert_check
        )
        error = self.client.get_error(response)
        if error:
            print('[{}] Failed to close test plan: "{}"'.format(TESTRAIL_PREFIX, error))
        else:
            print(
                "[{}] Test plan with ID={} was closed".format(
                    TESTRAIL_PREFIX, self.testplan_id
                )
            )

    def is_testrun_available(self) -> bool:
        """
        Ask if testrun is available in TestRail.

        :return: True if testrun exists AND is open
        """
        response = self.client.send_get(
            GET_TESTRUN_URL.format(self.testrun_id), cert_check=self.cert_check
        )
        error = self.client.get_error(response)
        if error:
            print(
                '[{}] Failed to retrieve testrun: "{}"'.format(TESTRAIL_PREFIX, error)
            )
            return False

        return response["is_completed"] is False

    def is_testplan_available(self) -> bool:
        """
        Ask if testplan is available in TestRail.

        :return: True if testplan exists AND is open
        """
        response = self.client.send_get(
            GET_TESTPLAN_URL.format(self.testplan_id), cert_check=self.cert_check
        )
        error = self.client.get_error(response)
        if error:
            print(
                '[{}] Failed to retrieve testplan: "{}"'.format(TESTRAIL_PREFIX, error)
            )
            return False

        return response["is_completed"] is False

    def get_available_testruns(self, plan_id: int) -> list[int]:
        """
        :return: a list of available testruns associated to a testplan in TestRail.

        """
        testruns_list = []
        response = self.client.send_get(
            GET_TESTPLAN_URL.format(plan_id), cert_check=self.cert_check
        )
        error = self.client.get_error(response)
        if error:
            print(
                '[{}] Failed to retrieve testplan: "{}"'.format(TESTRAIL_PREFIX, error)
            )
        else:
            for entry in response["entries"]:
                for run in entry["runs"]:
                    if not run["is_completed"]:
                        testruns_list.append(run["id"])
        return testruns_list

    def get_tests(self, run_id: int) -> Optional[list[dict]]:
        """
        :return: the list of tests containing in a testrun.
        """
        response = self.client.send_get(
            GET_TESTS_URL.format(run_id), cert_check=self.cert_check
        )
        error = self.client.get_error(response)
        if error:
            print('[{}] Failed to get tests: "{}"'.format(TESTRAIL_PREFIX, error))
            return None
        return response



from setuptools import setup


def read_file(fname):
    with open(fname) as f:
        return f.read()


setup(
    name='pytest_testrail_results',
    description='A pytest plugin to upload results to TestRail.',
    long_description=read_file('README.rst'),
    version='1.0.2',
    author='Oleg_Miroshnychenko',
    author_email='miroshnychenko@dnt-lab.com',
    packages=[
        'pytest_testrail_results',
    ],
    package_dir={'': 'src'},
    install_requires=[
        'pytest>=7.2.0',
        'requests>=2.30.0',
    ],
    include_package_data=True,
    entry_points={'pytest11': ['pytest-testrail = pytest_testrail_results.conftest']},
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
